/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <snort.h>
#include <s_util.h>

/*
* Snort parsing is now a 2 step process.
*
* 1. Build a list of key/value pairs (for our matchers). Activating
*    hacks as we go.
* 3. Actually build and add the rule(s), then cleanup the key/val
*    list.
*/

PLUGIN_DECLARE_STD_IMPORTS;

proc_add_rule		add_rule;
proc_add_criteria	add_criteria;
proc_set_target		set_target;
proc_add_parser		add_parser;
proc_rule_cleanup	rule_cleanup;

proc_list_append	list_append;
proc_list_remove	list_remove;

/* Hacks to make some keys target options */
char *opt_hacks[]={
	"DEPTH",
	"NOCASE",
	"OFFSET",
	NULL
};

/* Hacks for stuff to discard */
char *discard_hacks[]={
	"REFERENCE",
	"CLASSTYPE",
	"SID",
	"REV",
	NULL
};

/* Kinda sucks */
char *tpriv_hacks[]={
	"MSG",
	"LOGTO",
	NULL
};

/* This global data isn't protected. that means
* this plugin is in NO WAY reentrant. */
p_kv temp=NULL;
p_kv opts=NULL;
p_kv sip=NULL;
p_kv dip=NULL;
p_kv dpt=NULL;
p_kv spt=NULL;
char *target=NULL;
char *t_priv=NULL;

void snort_flip_rule(p_kv t)
{
	if ( sip && !strcasecmp("IP_SRC", sip->key) ) {
		strcpy(sip->key, "IP_DST");
	}
	if ( dip && !strcasecmp("IP_DST", dip->key) ) {
		strcpy(dip->key, "IP_SRC");
	}
	if ( spt && !strcasecmp("SPT", spt->key) ) {
		strcpy(spt->key, "DPT");
	}
	if ( dpt && !strcasecmp("DPT", dpt->key) ) {
		strcpy(dpt->key, "SPT");
	}
}

int snort_add_rule(void)
{
	p_kv p, the_opts;
	p_rule r;
	
	char *real_t_priv;
	
	if ( !t_priv )
		real_t_priv="SNORT";
	else
		real_t_priv=t_priv;
	
	if ( !(r=my_malloc(sizeof(rule_t))) ) return 0;
	SAFE_INIT_RULE_PTR(r);

	/* Set the rules target */
	if ( !(set_target(r, target, real_t_priv)) ) {
		rule_cleanup(r);
		return 0;
	}
	
	for (p=temp; p; p=(p_kv)p->l.next) {
		if ( !strcasecmp("CONTENT", p->key) ) {
			the_opts=opts;
		}else{
			the_opts=NULL;
		}
	
		if ( !add_criteria(r, p->key, p->val, the_opts, p->flags) ) {
			rule_cleanup(r);
			return 0;
		}
	}
	
	if ( !add_rule(r) ) {
		rule_cleanup(r);
		return 0;
	}
	
	return 1;
}


void snort_kv_clean(p_kv k)
{
	p_kv foo=NULL, bar=NULL;

	for (foo=k; foo;) {
		bar=foo;
		foo=(p_kv)foo->l.next;

		if ( bar->key ) my_free(bar->key);
		if ( bar->val ) my_free(bar->val);

		my_free(bar);
	}
}

void snort_cleanup_internal(void)
{
	p_kv foo=NULL, bar=NULL;
	
	if ( target ) my_free(target);
	if ( t_priv ) my_free(t_priv);

	snort_kv_clean(opts);
	snort_kv_clean(temp);
	
	target=t_priv=temp=opts=sip=dip=dpt=spt=NULL;
}

p_kv snort_link_element(const char *key, const char *val, int nval)
{
	p_kv n=NULL;
	int i;
	
	if ( !key ) return NULL;

	/* We can discard some entries right away based on key */
	for(i=0; discard_hacks[i]; i++) {
		if ( !strcasecmp(key, discard_hacks[i]) ) {
			return NULL;
		}
	}

	if ( !(n=my_malloc(sizeof(kv_t))) ) return NULL;

	if ( nval == NVAL_AUTO ) {
		if ( *val == '!' ) {
			val++;
			n->flags=1;
		}else{
			n->flags=0;
		}
	}else{
		n->flags = (nval) ? 1 : 0 ;
	}
	
	if ( !(n->key=my_strdup(key)) ) {
		my_free(n);
		return NULL;
	}

	
	if ( val ) {
		if ( !(n->val=my_strdup(val)) ) {
			my_free(n->key);
			my_free(n);
			return NULL;
		}
	}else{
		n->val = NULL;
	}

	for(i=0; tpriv_hacks[i]; i++) {
		if ( !strcasecmp(n->key, tpriv_hacks[i]) ) {
			if ( t_priv == NULL ) {
				t_priv=n->val;
			}else{
				my_free(n->val);
			}
			my_free(n->key);
			my_free(n);
			return NULL;
		}
	}

	for(i=0; opt_hacks[i]; i++) {
		if ( !strcasecmp(n->key, opt_hacks[i]) ) {
			opts=list_append(n, opts);
			return n;
		}
	}
	
	temp=list_append(n, temp);
	return n;
}

/* TODO:
* Allow the escaping of quote marks inside quoted strings
* */
#define PARSE_CRAP	0
#define PARSE_KEY	1
#define PARSE_VAL	2
int snort_parse_body(char *body)
{
	char *cur=NULL;
	int  cur_meaning=PARSE_CRAP;
	char *begin_key=NULL;
	char *begin_val=NULL;
	u_int32_t keys=0;
	int quote=0;
	int afterwhite=0;
	char *end;

	end=body+strlen(body);

	for ( cur=body; cur <= end; cur++ ) {
		switch ( cur_meaning ) {
			case PARSE_CRAP:
				if ( quote ) break;
				if ( *cur == '(' ) break;
				if ( *cur == ')' ) break;
				if ( !isblank(*cur) ) {
					cur_meaning=PARSE_KEY;
					begin_key=cur;
				}
				break;
			case PARSE_KEY:
				if ( quote ) break;
				if ( *cur == ':' ) {
					*cur='\0';
					cur_meaning=PARSE_VAL;
					begin_val=cur+1;
				}else if ( *cur == ';' ) {
					*cur='\0';
					snort_link_element(begin_key, NULL, NVAL_NORMAL);
					cur_meaning=PARSE_CRAP;
					quote=0;
					begin_val=NULL;
					begin_key=NULL;
				}
				break;
			case PARSE_VAL:
				if ( !afterwhite && isblank(*cur) ) {
					begin_val=cur+1;
					break;
				}
				afterwhite=1;
				if ( *cur == '"' ) {
					if ( !quote ) begin_val=cur+1;

					if ( quote ) *cur='\0';
					
					quote=(quote) ? 0 : 1;
				}
				if ( quote ) break;
				if ( *cur == ';' || *cur==')') {
					*cur='\0';

					/* FINISHED */
					snort_link_element(begin_key, 
						begin_val, NVAL_NORMAL);
					keys++;
					cur_meaning=PARSE_CRAP;
					quote=0;
					afterwhite=0;
					begin_val=NULL;
					begin_key=NULL;
				}
				break;
		}
	}
	
	return keys ? 1 : 0;
}

snort_ret snort_parse_rule(char *rule)
{
	p_rule nrule;
	char *tokens[5];
	int i=0;
	char *work;
	snort_ret ret=SNORT_NULL;
	char *vr;
	
	target=my_strdup(strtok(rule," \t"));
	if ( !target ) return SNORT_NULL;
	
	/* Get the head part tokenised:
	* target src spt -> dst dpt */
	while ( (tokens[i]=strtok(NULL," \t"))  && (i++<5) );
	if ( i != 6 ) return SNORT_NULL;
	
	/* Which way the rule goes, back, forth, bi */
	if ( !strcmp(tokens[3], "<-") ) { ret=SNORT_BACKWARDS; }
	if ( !strcmp(tokens[3], "->") ) { ret=SNORT_FORWARDS; }
	if ( !strcmp(tokens[3], "<>") ) { ret=SNORT_BIDIRECTIONAL; }
	
	/* Has to go back OR forth */
	if ( ret == SNORT_NULL ) return ret;
	
	/* Add the head stuff to the rule */
	if ( strcasecmp("any", tokens[0]) && strcasecmp("ip", tokens[0]))
		snort_link_element("IP_PROTO", tokens[0], NVAL_NORMAL);

	if ( tokens[1][0] == '$' ) {
		if ( (vr=get_var(tokens[1]+1)) )
			tokens[1]=vr;
		else
			print_out("SNORT VARIABLE %s NOT FOUND" CRLF,
				tokens[1]);
	}

	if ( tokens[4][0] == '$' ) {
		if ( (vr=get_var(tokens[4]+1)) )
			tokens[4]=vr;
		else
			print_out("SNORT VARIABLE %s NOT FOUND" CRLF,
				tokens[4]);
	}
	
	if ( strcasecmp("any", tokens[1]) )
		sip=snort_link_element("IP_SRC", tokens[1], NVAL_AUTO);
	if ( strcasecmp("any", tokens[2]) ) 
		spt=snort_link_element("SPT", tokens[2], NVAL_AUTO);
	if ( strcasecmp("any", tokens[4]) ) 
		dip=snort_link_element("IP_DST", tokens[4], NVAL_AUTO);
	if ( strcasecmp("any", tokens[5]) ) 
		dpt=snort_link_element("DPT", tokens[5], NVAL_AUTO);
	

	/* We now have the variable sized rule body
	* in this format:
	* (key:123; key2:"val"; key:"val |0d|";)  */
	
	if ( !(work=my_strdup(strtok(NULL, ""))) ) {
		return SNORT_NULL;
	}
	
	if ( !snort_parse_body(work) ) ret = SNORT_NULL;
	
	my_free(work);
	return ret;
}

int snort_loadfile(char *fn)
{
	FILE *f;
	char buf[4096];
	char *p=(char *)&buf;
	int rule=0, line=0;
	snort_ret ret;
	
	if ( !fn ) return 0;
	
	if ( (f=fopen(fn, "r")) == NULL ) {
		print_out(PRINT_ERROR "Unable to open snort rulefile: %s" CRLF, fn);
		return 0;
	}

	print_out("Opening snort rulefile (%s)" CRLF, fn);

	while( fgets(p, 4096, f) ) {
		line++;
		if ( IS_RULE(buf) ) {
			rule++;
			ret = snort_parse_rule(p);
			
			if ( ret == SNORT_NULL ) {
				print_out("Parse error on rule %i (%s:%i)"
					CRLF, rule, fn, line);
				goto rule_dud;
			}

			if ( ret == SNORT_FORWARDS || 
				ret == SNORT_BIDIRECTIONAL ) {
				if ( !snort_add_rule() ) {
					print_out("Rule error on rule %i "
						"(%s:%i)" CRLF, 
						rule, fn, line);
					goto rule_dud;
				}
			}
			
			if ( ret == SNORT_BACKWARDS || 
				ret == SNORT_BIDIRECTIONAL ) {
				snort_flip_rule(temp);
				if ( !snort_add_rule() ) {
					print_out("Rule error on backwards rule %i "
						"(%s:%i)" CRLF,
						rule, fn, line);
					goto rule_dud;
				}
			}

rule_dud:
			/* Clean up our parsers little scratch area */
			snort_cleanup_internal();
		}
	}
	
	fclose(f);
	return 1;
}

uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	PLUGIN_CHECK_PARAMS(in);
	PLUGIN_IMPORT_STD;
	
	/* Take the functions we need */
	add_rule=in->add_rule;
	add_criteria=in->add_criteria;
	set_target=in->set_target;
	add_parser=in->add_parser;
	rule_cleanup=in->rule_cleanup;
	list_append=in->list_append;
	list_remove=in->list_remove;
	
	/* Fill in a description about ourselves */
	PLUGIN_NAME("ruleset/snort");
	PLUGIN_DESC("Loads snort ruleset files");
	PLUGIN_VERS(1,1);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	
	/* Tell firestorm what function to
	* call when it needs to load a 'Snort' file */
	add_parser("Snort", snort_loadfile);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	return PLUGIN_ERR_OK;
}
